/**************************************************************************

    This file is part of 'tfc for windows', a character-mode graphical
	side-by-side file compare utility.

    Copyright (c) 1996 - 2000 by Daniel Schroeder (tadzio@tadzio.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation (see gpl.txt in this package); 
	either version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License (gpl.txt) for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

***************************************************************************/


/**************************************************************************
***************************************************************************
**
** MODULE      compare.c                              (c) D.Schrder 96-97
**
** PROJECT
**
** COMPILER    MS Visual C++ 5.0
**
** AUTHOR      Daniel Schrder
**
** ABSTRACT
**
**
** HISTORY     21.6.97 Portierung auf Win32
**             7.9.96 14:31 isBlankLine() und createIndex() behandeln jetzt
**               auch NL als Zeilenendekennzeichen
**             8.9.96 12:17 alle Funktionen benutzen jetzt keine Endekennung
**               mehr, sondern Lngenangaben. Dadurch ist ein Binrvergleichs-
**               modus mglich!
**
***************************************************************************
**************************************************************************/

/* INCLUDE FILES *********************************************************/

/* IMPORT */
#include <windows.h>
#include <stdlib.h>
#include <ctype.h>
#include "option.h"
#include "utility.h"

/* EXPORT */
#include "compare.h"

/* GLOBAL DEFINES ********************************************************/

// mgliche Werte (bitcodiert?) fr properties (Ma der bereinstimmung zweier Zeilen)
#define DIFFERENT   0
#define IDENTICAL   1

/* GLOBAL VARIABLES ******************************************************/

/* GLOBAL CONSTANTS ******************************************************/
#ifdef HEADER

struct _line {             // Beschreibung einer Zeile:
  PUCHAR pointer;          // Zeiger auf Speicherbereich
  UINT length;             // Lnge der Zeile
  };

struct _combined {
  PUCHAR line0;            // Zeiger auf Zeile (String; nicht terminiert!)
  UINT length0;            // Lnge der Zeile
  UINT lineno0;            // Nummer der Zeile im Originalfile
  PUCHAR line1;
  UINT length1;
  UINT lineno1;
  UINT properties;         // Flag fr Zeilen identisch/nicht identisch
  };

#endif

/* LOCAL DEFINES ********************************************************/
#define DEFAULT_LINES   4        // Erklrung siehe LineMatcher()
#define LO_RANGE      100
#define HI_RANGE      600

/* LOCAL VARIABLES *******************************************************/

/* LOCAL CONSTANTS *******************************************************/
static const PUCHAR emptyLine = "";     /* Darstellung einer leeren Zeile */

/* LOCAL FUNCTIONS *******************************************************/

/* FUNCTION **************************************************************/

  __inline int isBlankLine(struct _line line)

/*-------------------------------------------------------------------------
SPECIFICATION: stellt fest, ob die bergebene Zeile leer ist. "Leer" bedeutet
               hier: ber die ganze in line.length bergebene Lnge hinweg
               beinhaltet die Zeile (line.pointer) nur Whitespace (Leerzeichen, Tab)
               Wird NULL bergeben, ist das Ergebnis FALSE.

               Aus Performance-Grnden ist die Funktion als __inline definiert

Parameters:    line: Struktur mit Zeiger auf zu untersuchende Zeile und Lnge.

Side Effects:  keine

Return Value:  TRUE:  die Zeile enthlt nur Whitespace
               FALSE: sonst

-------------------------------------------------------------------------*/
{
  UINT i;

  for(i=0; i < line.length; i++)
    if(!isspace(line.pointer[i])) return(FALSE);

  return(TRUE);
}


/* FUNCTION **************************************************************/

  __inline UINT linelength(PUCHAR pointer, int maxlength, PUINT pSlack, int mode)

/*-------------------------------------------------------------------------
SPECIFICATION: Fat 'pointer' als Zeiger auf eine Zeile auf und ermittelt
               deren Lnge. Der verwendete Algorithmus hngt ab von mode:
               mode == 0: Zeile wird begrenzt durch '\0'
               mode == 1: Zeile wird begrenzt durch CR, LF oder CR/LF
               mode < 0:  Zeile hat feste Lnge = -mode.
               Bei mode >= 0 gehren der oder die Terminatoren NICHT mit
               zur Zeile, sondern zu slack.
               Die Lnge kann nie grer als maxlength werden.
               Es wird garantiert, da die Funktion nicht versucht, auf
               Speicher hinter pointer+maxlength zuzugreifen.

Parameters:    pointer: Zeiger auf den zu untersuchenden Speicherbereich
               maxlength: Maximale Lnge der Zeile
               pSlack: Zeiger auf Speicher, in dem die Anzahl der Zeichen
                 "zwischen den Zeilen" zurckgegeben wird.
               int mode: Stellt den verwendeten Algorithmus ein, s.o.

Side Effects:  keine

Return Value:  ermittelte Zeilenlnge

-------------------------------------------------------------------------*/
{
  int length;

  if(maxlength <= 0)
    {
    *pSlack = 0;
    return(0);
    }

  if(mode == 0)
    {
    *pSlack = 1;
    for(length=0; length < maxlength; length++)
      if(pointer[length] == '\0') break;
    }

  else if(mode == 1)
    {
    *pSlack = 1;
    for(length=0; length < maxlength; length++)
      {
      if(pointer[length] == NL) break;

      if(pointer[length] == CR)
        {
        if(length < maxlength-1)
          if(pointer[length+1] == NL) *pSlack = 2;

        break;
        }
      }
    }

  else if(mode < 0)
    {
    *pSlack = 0;
    length = min(-mode, maxlength);
    }

  else
    {
    *pSlack = length = 0;
    }

  return(length);
}


/* FUNCTION **************************************************************/

static int tfc_strcmp(struct _line p0, struct _line p1, int caseInsensitive, int skipWhitespace)

/*-------------------------------------------------------------------------
SPECIFICATION:  Vergleicht die beiden Zeichenketten p0 und p1 auf
                Gleichheit. "Gleichheit" ist hier wie folgt zu verstehen:
                Jeweils zwei Zeichen mssen identisch sein. Bei mehr als einem
                aufeinanderfolgenden Whitespace werden alle bis auf das letzte
                Whitespace ignoriert, sofern 'skipWhitespace' gesetzt ist.
                Da jeweils zwei Zeichen korrespondieren mssen, mu auch die
                Lnge (ggf vermindert um die Anzahl bersprungener Whitespace)
                beider Zeilen identisch sein.
                Ist caseInsensitive ungleich 0, wird nicht zwischen Gro- und
                Kleinschreibung unterschieden (dazu wird eine eigene Variante
                von tolower() verwendet, die auch deutsche Umlaute
                bercksichtigt).

Parameters:     p0, p1: Strukturen, die die zu vergleichende Zeichenketten
                  beschreiben. An sie werden keine
                  besonderen Bedingungen gestellt, sie drfen '\0' enthalten
                  und mssen nicht terminiert sein.

Side Effects:   keine

Return Value:   IDENTICAL, wenn die Zeichenketten gleich sind.
                DIFFERENT, wenn sie sich unterscheiden oder mindestens eine
                  von beiden NULL ist.

-------------------------------------------------------------------------*/
{
  unsigned int i0=0, i1 = 0;  // Indices in die beiden Zeilen

  if(p0.pointer == NULL || p1.pointer == NULL)
    return(DIFFERENT);


  while(i0 < p0.length && i1 < p1.length)
    {
    // Wenn mehr als ein Whitespace hintereinander kommt: berlesen, falls gewnscht
    if(skipWhitespace)
      {
      while(isspace(p0.pointer[i0]) && (i0 < p0.length-1) && isspace(p0.pointer[i0+1]))
        i0++;

      while(isspace(p1.pointer[i1]) && (i1 < p1.length-1) && isspace(p1.pointer[i1+1]))
        i1++;
      }

    // ------------------ Zeichen vergleichen ------------------------
    if(caseInsensitive)
      {
      if(xtolower(p0.pointer[i0]) != xtolower(p1.pointer[i1])) return(DIFFERENT);
      }
      
    else
      {
      if(p0.pointer[i0] != p1.pointer[i1])  return(DIFFERENT);
      }

    i0++; i1++;   // Auf geht's zum nchsten Zeichen!
    }


  /* Hierher kommen wir nur, wenn mindestens eine Zeile komplett durchlaufen
  wurde. Es kann sein, da ab dem momentanen Index in der anderen Zeile nur noch
  Whitespaces stehen. Diese berlesen wir jetzt, falls gewnscht. */
  if(skipWhitespace)
    {
    while(i0 < p0.length)
      {
      if(!isspace(p0.pointer[i0])) break;
      i0++;
      }

    while(i1 < p1.length)
      {
      if(!isspace(p1.pointer[i1])) break;
      i1++;
      }
    }


  // Gleich sind beide Zeilen genau dann, wenn wir beide bis zum Ende durcharbeiten konnten!
  return((i0 == p0.length && i1 == p1.length)? IDENTICAL:DIFFERENT);
}


/* FUNCTION **************************************************************/

 __inline void setCombined(struct _combined *pcombined, PUCHAR line0, UINT length0, UINT lineno0, PUCHAR line1, UINT length1, UINT lineno1, UINT properties)

/*-------------------------------------------------------------------------
SPECIFICATION: Hilfsfunktion; trgt die bergebenen Werte line0, length0,
               lineno0, line1, length1, lineno1 und properties in die per
               Zeiger bergebene struct _combined ein.
               Dient nur der bersichtlicheren Darstellung.

               Aus Performance-Grnden ist die Funktion als _Inline definiert

Parameters:    Zeiger auf Ziel-Struktur und diverse Quellwerte

Side Effects:  Die Struktur, auf die pcombined zeigt, wird beschrieben

Return Value:  keine

-------------------------------------------------------------------------*/
{
  pcombined -> line0 = line0;
  pcombined -> length0 = length0;
  pcombined -> lineno0 = lineno0;
  pcombined -> line1 = line1;
  pcombined -> length1 = length1;
  pcombined -> lineno1 = lineno1;
  pcombined -> properties = properties;

  return;
}


/* FUNCTION **************************************************************/

  __inline int miniLineMatcher(struct _line *strings0, struct _line *strings1, UINT lines)


/*-------------------------------------------------------------------------
SPECIFICATION:  Vergleicht die nchsten 'lines' Zeilen von strings0 mit
                strings1 und gibt eine Mazahl fr die bereinstimmung
                zurck. Die Mazahl ist das akkumulierte Ergebnis aller
                tfc_strcmp()-Aufrufe, welches fr jede Zeile genau einmal
                aufgerufen wird. Die Funktion ist eine Hilfsfunktion des
                "groen" LineMatchers, der diese Vergleiche in zwei
                Schleifen ausfhrt.
                Wird im Laufe der Schleife einer der beiden 'lines.pointer'-Werte
                NULL, bricht die Funktion den Vergleich ab und gibt das bisherige
                Ergebnis zurck.
                Aus Performance-Grnden ist die Funktion als _Inline definiert
                (was sie automatisch auch static macht);
                Achtung, dies wirkt nur, wenn die Compileroption /Oi+ verwendet
                wird!

Parameters:    strings0, strings1: Zwei Felder von Strukturen, die die zu vergleichenden
                 Zeilen beschreibt. strings0.pointer und strings1.pointer
                 mssen einen NULL-Eintrag als Endekennzeichen haben.

               lines: Anzahl zu vergleichender Zeilen

Side Effects:

Return Value:

-------------------------------------------------------------------------*/
{
  UINT i, score = 0;

  for(i=0; i < lines; i++)
    {
    if(strings0[i].pointer == NULL || strings1[i].pointer == NULL)
      break;
    score += tfc_strcmp(strings0[i], strings1[i], options.caseInsensitive, !options.binmode && options.compresspace);
    }

  return(score);
}


/* FUNCTION **************************************************************/

 static UINT LineMatcher(struct _line *strings0, struct _line *strings1,
   PUINT pi, PUINT pj, UINT lines, UINT range)

/*-------------------------------------------------------------------------
SPECIFICATION: Die Funktion vergleicht, ausgehend von den bergebenen
               Werten in pi und pj, die Zeilen, auf die strings0[*pi] und
               strings1[*pj] zeigen, und findet das am besten passende
               Prchen, wobei *pi oder *pj hochgezhlt werden kann.
               Dabei wird folgender Algorithmus verwendet:
               Ist entweder strings0[*pi].pointer oder strings1[*pj].pointer NULL, wird der
               jeweils andere Zhler soweit hochgezhlt, bis sein zugehriger
               strings.pointer-Wert ebenfalls NULL ist; die Funktion wird dann beendet.
               Stimmen strings0[*pi] und strings1[*pj] berein, oder ist options.binmode
               gesetzt, bleiben *pi und *pj unverndert. Stimmen sie nicht berein, wird
               zunchst die beste bereinstimmung von 'lines' Zeilen aus
               strings0 (beginnend bei dem festen *pi) mit 'lines' Zeilen
               aus strings1 ermittelt, wobei in einer Schleife *pj incrementiert
               wird, bis es 'range' Zeilen vom Ausgangswert entfernt ist.
               "Beste bereinstimmung" bedeutet: mglichst viele der 'lines'
               Zeilen aus strings0 sind identisch mit den gerade untersuchten
               'lines' Zeilen aus strings1."Identisch" wird definiert durch
               die Funktion tfc_strcmp() - siehe dort.
               Bei Gleichheit der Anzahl bereinstimmender Zeilen ist der
               Bereich mit dem niedrigeren *pj besser.
               Dasselbe passiert nun mit vertauschten Rollen von strings0 und
               strings1 bzw pi und pj. Das insgesamt beste Ergebnis beider
               Suchschleifen (wenn beide Schleifen dasselbe Ergebnis liefern,
               ist das Ergebnis der Schleife besser, die weniger Zeilen berspringen
               mu) definiert neue Werte fr *pi und *pj (wobei sich
               nur *einer* der beiden Werte ndern kann!). Diese Werte werden
               in pi und pj eingetragen. Zustzlich gibt die Funktion noch
               das Ergebnis von tfc-strcmp() fr die beiden gerade gefundenen
               Zeilen zurck; dieses Ergebnis zeigt an, ob die beiden Zeilen
               identisch (nach der Definition von tfc-strcmp()) sind.

Parameters:    strings0, strings1: Zwei Felder von Strukturen, die die zu vergleichenden
                 Zeilen beschreiben; strings0 und strings1
                 mssen einen NULL-Eintrag als Endekennzeichen haben.

               pi, pj: Zeiger auf die zur Indizierung von strings0, strings1 verwendeten
                 Variablen. Beim Aufruf von LineMatcher() mssen sie mit sinnvollen
                 Werten vorbelegt sein. Die Suche startet bei strings0[*pi] bzw
                 strings1[*pj].
               lines: Blockgre, die verglichen werden soll; siehe SPECIFICATION
               range: Suchbereich, innerhalb dessen 'lines' Zeilen gefunden werden sollen.

Side Effects:  Liest die globale Variable options.binmode

Return Value:  Wert, der das Ma der bereinstimmung der beiden gefundenen
               Zeilen angibt. Zur Definition des Maes siehe tfc_strcmp()
               Zudem wird ggf der Inhalt von pi oder von pj auf den ermittelten
               Wert gesetzt.

-------------------------------------------------------------------------*/
{
  UINT i, j,                   // Laufindices fr Vergleichsoperationen
       hii = *pi, hij = *pj;   // Indices, bei denen 'hiscore' aufgetreten ist
  int hiscore = -1,            // bisher beste bereinstimmung eines Blockes; wird bei der ersten Zeile auf 0 steigen
       score;                  // Ergebnis des Vergleichs von 'lines' Zeilen


  if(strings0[*pi].pointer == NULL)
    {
    while(strings1[*pj].pointer != NULL)  (*pj)++;
    return(DIFFERENT);
    }

  if(strings1[*pj].pointer == NULL)
    {
    while(strings0[*pi].pointer != NULL)  (*pi)++;
    return(DIFFERENT);
    }

  if(options.binmode == TRUE)
    return(miniLineMatcher(strings0 + *pi, strings1 + *pj, 1));

  if(miniLineMatcher(strings0 + *pi, strings1 + *pj, 1) == IDENTICAL)
    return(IDENTICAL);

  // -------------- Erste Schleife: festes *pi, gleitendes j -----------------------
  for(j = *pj; (j < *pj+range) && (strings1[j].pointer != NULL); j++)
    {
    score = miniLineMatcher(strings0 + *pi, strings1 + j, lines);   // 'lines' Zeilen ab *pi und j vergleichen
    // tprintf("Loop 1: Comparing lines %4i and %4i; score %i; hiscore %i\n", *pi, j, score, hiscore);
    if(score > hiscore)
      {
      hiscore = score;
      hij = j;
      if(hiscore == (int)lines)               // Besser kann's nicht mehr werden, also abbrechen
        break;
      }
    }

  // -------------- Zweite Schleife: festes *pj, gleitendes i ----------------------
  for(i = *pi; i < *pi+range && strings0[i].pointer != NULL; i++)
    {
    if(hiscore == (int)lines && (i-*pi > hij-*pj))         // erste Schleife hat max. hiscore geschafft, und zweite Schleife hat schon grere Zahl bersprungener Zeilen, kann also nicht mehr besser sein -> Abbruch
      break;

    score = miniLineMatcher(strings0 + i, strings1 + *pj, lines);   // hier 'lines' Zeilen ab *pi und j vergleichen
    // tprintf("Loop 2: Comparing lines %4i and %4i; score %i; hiscore %i\n", i, *pj, score, hiscore);
    if( score > hiscore || ( score == hiscore && (i-*pi < hij-*pj)))
      {
      hiscore = score;
      hii = i;
      hij = *pj;                         // Ergebnis der ersten Schleife verwerfen, denn wir haben hier was besseres!
      if(hiscore == (int)lines)          // Besser kann's nicht mehr werden, also abbrechen
        break;
      }
    }

  *pi = hii;
  *pj = hij;

  return(miniLineMatcher(strings0 + *pi, strings1 + *pj, 1));
}


/* GLOBAL FUNCTIONS ******************************************************/

/* FUNCTION **************************************************************/

struct _line *createIndex(PUCHAR text, UINT textlength, int mode)

/*-------------------------------------------------------------------------
SPECIFICATION:  Erzeugt ein Feld von Strukturen, in dem jeder Eintrag auf eine
                neue Zeile in 'text' zeigt und die Lnge dieser Zeile enthlt
                Der bentigte Speicher wird
                dynamisch belegt, der Aufrufer ist fr seine Freigabe verant-
                wortlich. "Neue Zeile" wird durch 'mode' bestimmt.
                Ist text == NULL oder textlength == 0, dann gibt die Funktion
                eine leere Liste (nur mit dem Endekennzeichen) zurck.

                Ist options.compressline gesetzt, aber nicht options.binmode, dann
                werden alle Zeilen, die ausschlielich whitespace enthalten, ignoriert.

                Das Ende des erzeugten Feldes wird durch eine NULL angezeigt.

Parameters:     text:       Zeiger auf den zu verarbeitenden Speicherbereich
                textlength: Lnge des zu untersuchenden Speicherbereichs
                mode:       Wird an linelength() durchgereicht, Erklrung siehe dort!

Side Effects:   'text' wird verndert; Hauptspeicher wird dynamisch belegt.

Return Value:   Zeiger auf das erzeugte Feld von Strukturen, oder NULL im
                Fehlerfall

-------------------------------------------------------------------------*/
{
  UINT i=0, j=0, indexlength = 0, // i ist Laufvariable in 'text', j ist Laufvariable in 'index'
    slack;
  struct _line *index;

  /* Speicherbedarf (Anzahl Zeilen, abhngig vom Zeilenlngenkriterium), mal
  sizeof(struct _line) bestimmen; sozusagen Blinddurchlauf ohne Abspeichern der Ergebnisse */
  if(mode < 0)
    indexlength = textlength/-mode + 1;

  else
    {
    while(i < textlength)
      {
      i += linelength(text+i, textlength - i, &slack, mode);
      i += slack;
      indexlength++;
      }
    }

  index = malloc(sizeof(struct _line) * (indexlength+1));  // +1 fr den abschlieenden NULL-Eintrag
  if(index == NULL) return(NULL);

  i = 0;
  while(i < textlength)
    {
    index[j].pointer = text+i;
    index[j].length = linelength(text+i, textlength - i, &slack, mode);
    i+= index[j++].length + slack;

    if(options.compressline && !options.binmode)    // Aufteilung in zwei ifs, um optimale Bearbeitungsreihenfolge sicherzustellen
      if(isBlankLine(index[j-1]))
        j--;                        // Gefundene Zeile wieder killen
    }

  index[j].pointer = NULL;
  index[j].length = 0;

  return(index);
}


/* FUNCTION **************************************************************/

struct _combined *createCombinedList(struct _line *index0, struct _line *index1, PUINT pcombinedlength)

/*-------------------------------------------------------------------------
SPECIFICATION:  Die eigentliche Vergleichsroutine; erzeugt aus den bergabe-
                parametern ein Feld von struct _combined - Eintrgen. Jeder
                Eintrag im Ergebnisfeld reprsentiert eine anzuzeigende
                Zeile und enthlt einen Zeiger auf eine Zeile aus text0,
                die Lnge der Zeile in text0, einen Zeiger auf eine Zeile aus
                text 1, die Lnge der Zeile aus text1 und ein 'properties'-
                Feld, das derzeit nur angibt, ob die beiden Zeilen identisch
                sind oder nicht.
                Die Funktion belegt dynamisch Hauptspeicher fr die zu erzeugende
                Liste. Der Aufrufer ist fr dessen Freigabe verantwortlich.
                Die resultierende Liste hat hchstens soviele Eintrge wie index0
                plus index1. Sie ist mit einem NULL-Eintrag terminiert.

Parameters:     index0, index1: Zwei Felder von Strukturen, die die Zeilen der
                  zugehrigen Speicherbereiche beschreiben.
                pcombinedlength: Zeiger auf eine Rckgabevariable, in die
                  die Funktion die Lnge (in Eintrgen, ohne den abschlieenden
                  Nulleintrag) der gerade erstellten Liste eintrgt. Der Inhalt
                  der Variablen beim Aufruf ist irrelevant.

Side Effects:   Hauptspeicher wird dynamisch belegt

Return Value:   Zeiger auf die erzeugte Liste

-------------------------------------------------------------------------*/
{
  struct _combined *combined;
  UINT i=0, j=0, k, next_i, next_j;
  UINT properties, lines, range;       // lines und range enthalten Werte, die durch die Kommandozeile beeinflubar sind

  if(index0 == NULL || index1 == NULL)
    return(NULL);

  while(index0[i].pointer != NULL)  i++;
  while(index1[j].pointer != NULL)  j++;

  combined = malloc( (i+j+1) * sizeof(struct _combined));

  if(combined != NULL)
    {
    /* Es laufen zwei Indexzhler (i fr index0, j fr index1) und ein Zhler (k) fr combined.
    Zuerst wird, ausgehend vom aktuellen Stand von i und j, das nchste zusammengehrende
    Zeilenprchen gesucht (das macht der "LineMatcher" - er setzt dabei i und j auf die neuen
    Werte, wobei er immer nur eine der beiden Variablen ndert!). Dabei kann es (mu aber
    nicht - selbst dann nicht, wenn die Zeilen unterschiedlich sind!) passieren, da i oder j
    ein paar Zeilen berspringt (Ist index0 oder index1 am Ende, liefert der LineMatcher auto-
    matisch ein Ergebnis, das alle restlichen Zeilen des anderen, noch nicht leeren Index
    berspringt.) Die ggf bersprungenen Zeilen (um das feststellen zu knnen, verndert der
    LineMatcher nicht i und j, sondern gesonderte Variablen next_i bzw next_j) werden in
    combined eingetragen (wobei i und j auf next_i bzw next_j hochgezhlt werden), die
    andere Seite wird mit Leerzeilen aufgefllt, k wird entsprechend hochgezhlt. Falls nicht
    NULL, werden danach werden die beiden Zeilen des gefundenen Zeilenprchens in combined
    eingetragen, und der nchste Schleifendurchlauf beginnt.
    Da die Zhlung der Zeilen in combined bei 1 beginnen soll, werden bei Zeilenprchen statt
    i und j i+1 bzw j+1 eingetragen. Leerzeilen werden durch lineno=0 markiert.
    Das ganze luft in einer Schleife, bis sowohl index0 als auch index1 komplett verarbeitet
    wurden. combined wird mit einem Nulleintrag abgeschlossen. */
    i = j = k = 0;

    lines = options.linecount? options.linecount : DEFAULT_LINES;  // siehe LineMatcher()
    range = options.sync? LO_RANGE : HI_RANGE;                     //      - " -

    while(index0[i].pointer != NULL || index1[j].pointer != NULL)
      {

      next_i = i;                       // damit der LineMatcher wei, ab wo er suchen soll
      next_j = j;

      properties = LineMatcher(index0, index1, &next_i, &next_j, lines, range); // findet das nchste zusammengehrende Zeilenprchen

      for(  ; i < next_i; i++)          // Schleife fr evtl in index0 bersprungene Zeilen
        setCombined(combined + k++, index0[i].pointer, index0[i].length, i+1, emptyLine, 0, 0, DIFFERENT);

      for(  ; j < next_j; j++)          // Schleife fr evtl in index1 bersprungene Zeilen
        setCombined(combined + k++, emptyLine, 0, 0, index1[j].pointer, index1[j].length, j+1, DIFFERENT);


      if(index0[i].pointer != NULL && index1[j].pointer != NULL)
        {
        setCombined(combined + k, index0[i].pointer, index0[i].length, i+1, index1[j].pointer, index1[j].length, j+1, properties);
        i++; j++; k++;
        }
      }

    setCombined(combined + k, NULL, 0, 0, NULL, 0, 0, 0);          // Endekennzeichen
    }

  *pcombinedlength = k;

  return(combined);
}
